/* February 2020, © Ruud Harmsen, but do with it whatever you see fit.
   See the comments in SlashExpander.h for a full explanation. */
#include "SlashExpander.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <wctype.h>

static char *se_nextchar (char *cp, char *endp);
static char *se_prevchar (char *cp, char *bgnp, char *endp);
static int   se_isletter (char *cp, char *endp);
static se_SubstrSpec **se_AddCluster (se_SubstrSpec **Places, int cluster);
static int se_AddVariant (se_SubstrSpec **Places, int cluster, int variant, se_SubstrSpec vardata);
static se_SubstrSpec *se_Permute (se_SubstrSpec **Places, int (*ResFunc) (se_SubstrSpec *Result),
       se_SubstrSpec *Result, int cluster);
static se_SubstrSpec *se_AllocateResult (se_SubstrSpec **Places);

/* See the function prototype in file SlashExpander.h for an explanation. */
se_SubstrSpec **se_AnalyseString (char *line, char *separator)
{
   char *sep = separator;
   se_SubstrSpec **Places = NULL;
   int cluster = 0;
   int variant = 0;
   char *clline; /* Start of line segment for two segments */
   size_t slen;  /* Length of currently handled line segments */
   char *el;     /* End of currently handled line segments */
   char *b1, *e1, *b2, *e2; /* Begin and end of detected variants */
   char *sl; /* Currently handled separator, like a slash */
   se_SubstrSpec vardata;

   if (!separator || !strlen(separator))
      sep = "/";

   for (e2 = clline = line;
      sl = strstr(clline, sep), sl;
      cluster += 2, variant = 0)
   { /* Cluster loop */
      slen = strlen(clline);
      el = clline + slen;

      for (b1 = sl <= clline ? sl : se_prevchar(sl, clline, el);
         b1 && b1 > clline && se_isletter(b1, el);
         b1 = se_prevchar(b1, clline, el))
         ;
      if (!se_isletter(b1, el) && strncmp(b1, sep, strlen(sep)) != 0)
      {
         b1 = se_nextchar(b1, el);
      }
      e1 = sl;

      Places = se_AddCluster(Places, cluster);
      vardata.sub = clline;
      vardata.len = b1 - clline;
      se_AddVariant(Places, cluster, variant, vardata);

      Places = se_AddCluster(Places, cluster + 1);
      vardata.sub = b1;
      vardata.len = e1 - b1;
      se_AddVariant(Places, cluster + 1, variant, vardata);

      for (variant++; sl; variant++)
      { /* Loop for second etc. variants */
         b2 = sl + strlen(sep);
         for (e2 = b2;
            *e2 && se_isletter(e2, el);
            e2 = se_nextchar(e2, el))
            ;

         vardata.sub = b2;
         vardata.len = e2 - b2;
         se_AddVariant(Places, cluster + 1, variant, vardata);

         if (strncmp(e2, sep, strlen(sep)) == 0)
            sl = e2;
         else
            sl = NULL;
      } /* End of variant loop */
      clline = e2;
   } /* End of cluster loop*/

   /* Final non-variant string */
   if (clline = e2, slen = strlen(clline))
   {
      Places = se_AddCluster(Places, cluster);
      variant = 0;
      vardata.sub = clline;
      vardata.len = slen;
      se_AddVariant(Places, cluster, variant, vardata);
   }

   #ifdef DEBUG
   fprintf(stdout, "Showing it\n");
   for (cluster = 0; Places[cluster] && cluster < 10; cluster++)
   {
      for (variant = 0; Places[cluster][variant].sub; variant++)
      {
         fprintf(stdout, "cluster=%d, variant=%d, len=%2d, str=%d %s\n",
           cluster, variant,
            Places[cluster][variant].len,
            (int)(Places[cluster][variant].sub - line),
            Places[cluster][variant].sub);
         fflush(stdout);
      }
   }
   #endif

   return Places;
}

/* See the function prototype in file SlashExpander.h for an explanation. */
se_SubstrSpec *se_MakeResults (se_SubstrSpec **Places, int (*ResFunc) (se_SubstrSpec *Result))
{
   se_SubstrSpec *Result = se_AllocateResult(Places);

   return se_Permute(Places, ResFunc, Result, -1);
}

void se_FreeAll (se_SubstrSpec **Places, se_SubstrSpec *Result)
{
   int i;

   for (i = 0; Places[i]; i++)
      free(Places[i]);

   free(Result);
}

/* === End of publically accessible functions.
       Start of file local static functions. === */

/* This logically corresponds to a simple cp++, if we were
   dealing with single byte characters only. The function
   however can also handle multibyte characters, for example
   if the locale specifies that the encoding is UTF8.
   The function returns a pointer, advanced by the number
   of bytes the current character occupied, so it then
   points to the next multibyte character (of NULL is the
   end of the string is reached).
 */
static char *se_nextchar (char *cp, char *endp)
{
   int mx = endp - cp;
   int rt;

   if (mx > MB_CUR_MAX)
      mx = MB_CUR_MAX;

   rt = mblen(cp, mx);

   return cp + (rt <= 0 ? 1 : rt);
}

/* This logically corresponds to a simple cp--, if we were
   dealing with single byte characters only. The function
   however can also handle multibyte characters, for example
   if the locale specifies that the encoding is UTF8.
   The function returns a pointer, moved back by the number
   of bytes the previous character occupies, so it then
   points to that previous multibyte character. The function
   will not go beyond the start of the string, so if cp and
   bgnp are equal, the returned pointer will be NULL to signal
   error.
 */
static char *se_prevchar (char *cp, char *bgnp, char *endp)
{
   int mx = endp - cp;
   int rt;

   if (mx > MB_CUR_MAX)
      mx = MB_CUR_MAX;
   for (--cp; cp >= bgnp; cp--)
   {
      rt = mblen(cp, mx);
      if (rt > 0)
         return cp;
   }
   return NULL;
}

/* This function converts the multibyte character under cp
   to a wide character, and tests if it is alphabetic or
   a dash '-'.
 */
static int se_isletter (char *cp, char *endp)
{
   int mx = endp - cp;
   int rt;
   wchar_t wc;

   if (*cp == '-')
      return 1;

   if (mx > MB_CUR_MAX)
      mx = MB_CUR_MAX;

   rt = mbtowc(&wc, cp, endp - cp);
   if (rt <= 0)
      return 0;

   return iswalpha(wc);
}

/* This adds a new cluster to the array pointed to by Places.
   If Places is NULL, the array is initially allocated.
   The new cluster contains to variants yet, signalled by a
   single NULL.
   Places gets a NULL after the new cluster, so it is known
   where Places currently ends.
 */
static se_SubstrSpec **se_AddCluster (se_SubstrSpec **Places, int cluster)
{
    se_SubstrSpec **NewPointer;

    if (!(NewPointer = realloc(Places, (cluster + 2) * sizeof (se_SubstrSpec *))))
       return NULL;

    Places = NewPointer;
    /* Ready for adding a variant. */
    Places[cluster] = NULL;
    /* End indicator */
    Places[cluster + 1] = NULL;

    return Places;
}

/* This adds a pointer to an variant struct, to the specified cluster in Places,
   and fills it with vardata.
   After that comes a new NULL, to indicate the end of that cluster.
 */
static int se_AddVariant (se_SubstrSpec **Places, int cluster, int variant, se_SubstrSpec vardata)
{
    se_SubstrSpec *NewPointer;

    if (!Places)
       return 1;

    if (!(NewPointer = realloc(Places[cluster], (variant + 1) * sizeof (se_SubstrSpec))))
       return 2;

    Places[cluster] = NewPointer;
    Places[cluster][variant] = vardata;
    Places[cluster][variant + 1].sub = NULL;

    return 0;
}

/* Recurse clusters and variants, until *ResFunc has been called once for each
   expanded result.
 */
static se_SubstrSpec *se_Permute (se_SubstrSpec **Places, int (*ResFunc) (se_SubstrSpec *Result),
   se_SubstrSpec *Result, int cluster)
{
   /* Recursive algorithm inspired by
      https://www.tutorialcup.com/string/recursively-print-sentences-formed-word-lists.htm */
   if (cluster < -1)
      return NULL;

   if (!Result)
      return NULL;

   if (++cluster, Places && Places[cluster] && Places[cluster][0].sub)
   {
      int variant;

      /* No stop condition applies, so continue recursion. */
      for (variant = 0, Result[cluster] = Places[cluster][variant];
           Places[cluster][variant].sub;
           variant++, Result[cluster] = Places[cluster][variant])
      {
         se_Permute(Places, ResFunc, Result, cluster);
      }
   }
   else /* Stop condition applies: deliver result */
   {
      (*ResFunc)(Result);
   }
   return Result;
}

/* Add as many array members for Result as, after se_AnalyseString
   has been called, there are clusters in Places. Terminate with
   a NULL. The function returns a pointer to the allocated buffer
   Result.
 */
static se_SubstrSpec *se_AllocateResult (se_SubstrSpec **Places)
{
   /* The array that Places points to has its definite size now,
      so that can be used to allocate a matching Result array. */
   int numelts;

   for (numelts = 0; Places[numelts]; numelts++)
      ;

   return calloc(numelts, sizeof (se_SubstrSpec));
}

/* Test driver */
#ifdef MAIN
#include <stdio.h>
#include <locale.h>
#include "SlashExpander.h"

int ShowResult (se_SubstrSpec *Result);

int main (int argc, char **argv)
{
   FILE *fpi = stdin, *fpo = stdout;
   static char InputLine[512];
   se_SubstrSpec **Places;
   se_SubstrSpec *Result;
   long linecnt = 0;

   /* Use the locale that was specified in environment variables */
   setlocale(LC_ALL, "");

   while(fgets(InputLine, sizeof InputLine, fpi))
   {
      linecnt++;
      if (!(Places = se_AnalyseString(InputLine, "/")))
      {
         fprintf(stderr, "%s: Progline=%3d, inputline=%6ld\n",
            argv[0], __LINE__, linecnt);
         fprintf(stderr, "%s: Cannot allocate Places buffer\n", argv[0]);
         return 1;
      }
      if (!(Result = se_MakeResults(Places, &ShowResult)))
      {
         fprintf(stderr, "%s: Progline=%3d, inputline=%6ld\n",
            argv[0], __LINE__, linecnt);
         fprintf(stderr, "%s: Cannot allocate Result buffer\n", argv[0]);
         return 2;
      }

      se_FreeAll(Places, Result);
   }

   return 0;
}

/* This function will be called by se_MakeResults, once for each expansion
   of the variants in the clusters. */
int ShowResult (se_SubstrSpec *Result)
{
   int cluster;

   for (cluster = 0; Result[cluster].sub; cluster++)
   {
      fprintf(stdout, "%.*s", Result[cluster].len, Result[cluster].sub);
   }

   return 0;
}
#endif
